<?php

/*
	name - represents a single name from an identity
	Copyright (c) 2009
	
	This file is part of NicNames.
	
	NicNames is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as 
	published by the Free Software Foundation, either version 3 of
	the License, or (at your option) any later version.
	
	NicNames is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with NicNames.  If not, see <http://www.gnu.org/licenses/>.
*/

require_once(RESOURCE_DIR . 'includes/relation.inc.php');
require_once(RESOURCE_DIR . 'includes/probability.inc.php');

class name extends relation
{
	public $id = null;
	public $surname;
	public $type;
	public $givennames = '';
	public $title = '';
	public $surnamefirst = false;

	function getfields()
	{
		return array_merge(array(
			array('type', 'select', 'type', 'Type of name', null, true, nicnames_config::$resourcetypes['name_type'], true),
			array('surname', 'string', 'surname', 'Enter the surname part of this name', 128, true, array('size' => 12)),
			array('givennames', 'string', 'given names', 'Enter the given names for this name (separated by spaces)', 255, false, array('size' => 24)),
			array('title', 'string', 'title', 'Enter a title, if appropriate (Dr, Prof, etc)', 64, false, array('size' => 4)),
			array('surnamefirst', 'yesno', 'surname goes first', 'Surname should appear before given names', null, true, array('canhide' => true)),
			), $this->getcommonfields());
	}

	function getXML($document = null)
	// Create a DomElement object representing this name
	{
		if ($document===null) $document = new DomDocument();
		$el = $document->createElement('name');
		$el->appendChild($document->createElement('surname', $this->surname));
		$el->setAttribute('type', $this->type);
		if ($this->givennames != '') $el->appendChild($document->createElement('givennames', $this->givennames));
		if ($this->title != '') $el->appendChild($document->createElement('title', $this->title));
		if ($this->surnamefirst) $el->setAttribute('surnamefirst', 'true');
		if ($this->id) $el->setAttribute('id', $this->id);
		$el = $this->getcommonpropertiesXML($el);
		return $el;
	}

	function setfromXML($el)
	// populate this object with properties from the supplied DOM XML element
	{
		if ($el->hasAttribute('id')) $this->id = $el->getAttribute('id');
				
		for ($i = 0, $len = $el->childNodes->length; $i < $len; $i++)
		{
			$node = $el->childNodes->item($i);
			switch ($node->nodeName)
			{
			case 'surname':
				$this->surname = $node->nodeValue;
				break;
			case 'givennames':
				$this->givennames = $node->nodeValue;
				break;
			case 'title':
				$this->title = $node->nodeValue;
				break;
			}
		}

		$this->type = $el->getAttribute('type');

		$this->setcommonfromXML($el);
	}

	function similarityto($other)
	{
		$factors = array();

		// surname and given names
		$names1 = preg_split('/[\s.,]++/S', trim($this->surname . ' ' . $this->givennames), -1, PREG_SPLIT_NO_EMPTY);
		$names2 = preg_split('/[\s.,]++/S', trim($other->surname . ' ' . $other->givennames), -1, PREG_SPLIT_NO_EMPTY);

		$fuzzy1 = $fuzzy2 = array();
		foreach ($names1 as $name) $fuzzy1[] = new textfuzzy($name);
		foreach ($names2 as $name) $fuzzy2[] = new textfuzzy($name);

		$rows = count($names1);
		$cols = count($names2);

		$scores = array();
		foreach ($fuzzy1 as $cmp1)
			foreach ($fuzzy2 as $cmp2)
		{
			$name1 = $cmp1->getnormal();
			$name2 = $cmp2->getnormal();
			if (strlen($name1) == 1 || strlen($name2) == 1)
				$scores[] = $name1[0] == $name2[0] ? 0.58 : 0.08; 
			else
				$scores[] = $cmp1->similarityto($cmp2);
		}

		asort($scores);

		$scorevals = array(48, 5, 3, 3, 4, 5, 6, 7, 8, 8, 9, 9, 10);

		// "madonna factor" - it is rare for someone to have only a single name, so
		// two people with a single name that matches is a pretty sure bet
		if ($rows <= 1 && $cols <= 1) $scorevals[0] = 160;

		// pair off the names with their closest matches (don't match one from either side more than once)
		for ($i = 0, $min = min($cols, $rows), $max = max($cols, $rows); $i < $max; $i++)
			$factors["name$i"] = array(0, $scorevals[$i] * ($i < $min ? 1 : 0.2));

		while (!empty($scores))
		{
			$score = end($scores);
			$key = key($scores);
			$row = floor($key / $cols);
			$col = $key - ($row * $cols);

			$max = max($row, $col);
			while (isset($factors["name$max"]) && $factors["name$max"][0] > 0) $max++;
			$factors["name$max"] = array($score, 				
				($row === $col ? 1 : 0.3)
				* (isset($scorevals[$max]) ? $scorevals[$max] : 1)
				* ($score < 0.5 ? 14 : 1));
			
			for ($i = 0; $i < $rows; $i++) 
				unset($scores[$i * $cols + $col]);
			for ($i = 0; $i < $cols; $i++)
				unset($scores[$row * $cols + $i]);
		}

		// title
		if ($this->title != '' || $other->title != '')
		{
			if ($this->title == '' || $other->title == '')
				$factors["title"] = array(0, 0.4);
			else
			{
				$title1 = new textfuzzy($this->title);
				$title2 = new textfuzzy($other->title);
				$factors["title"] = array($title1->similarityto($title2), 3);
			}
		}

		$factors['type'] = array($this->type == $other->type, 2);

		if ($this->surnamefirst || $other->surnamefirst)
			$factors['surnamefirst'] = array($this->surnamefirst == $other->surnamefirst, 2);

		$factors['common'] = $this->commonsimilarityfactor($other);

		return probability::combinedcorrected($factors);
	}

	function getstrings()
		// returns array of strings suitable for human-readable rendering of this
		// piece of informtion.  Contains such fields as 'title', 'subtitle', 
		// 'pre-qualifier', 'post-qualifier', 'comment', etc
	{
		$types = nicnames_config::$resourcetypes;

		$type = isset(nicnames_config::$resourcetypes['name_type'][$this->type]) ?
			nicnames_config::$resourcetypes['name_type'][$this->type] : $this->type;
		$givens = $this->givennames == '' ? null : $this->givennames;
		return array(
			'title' => $this->surname,
			'subtitle' => $givens,
			'pre-qualifier' => $type,
			'post-qualifier' => $this->title == '' ? null : ('(' . $this->title . ')'),
			'comma' => !$this->surnamefirst && $givens != '',
		) + $this->getcommonstrings();
	}
		
/*

		// surname
		$dif = $this->comparestrings($this->surname, $parts2['surname']);
		$bon = in_array($parts2['surname'], $this->givennames)
			||   in_array($this->surname, $parts2['givennames']);
		$score = $bon ? min(40, $dif) : $dif;		
		
		// givennames
		$given1 = $this->givennames;
		$given2 = $parts2['givennames'];
		$scorevals = array(50, 30, 10, 8, 6, 4, 2, 1);
		while (!empty($scorevals))
		{
			$name1 = array_shift($given1);
			$name2 = array_shift($given2);
			$scoreval = array_shift($scorevals);
			$nscore = 0;
			if (!strlen($name1) || !strlen($name2)) $nscore = 35;
			else
			{
				$initials = (preg_match('#\.$#', $name1) || preg_match('#\.$#', $name2));				if (!$initials)
					$nscore = $this->comparestrings($name1, $name2);
				else
					$nscore = 20;	
				if ($name1{0} != $name2{0}) $nscore += 50;
			}
			$score += round(($nscore * $scoreval) / 100);
		}
		
		// title
		if ($this->title != $parts2['title'])
		{
			if (!strlen($this->title) || !strlen($parts2['title'])) $score += 2;
				else $score += 6;
		}
		
		// extra
		if ($this->extra != $parts2['extra'])
		{
			$extra1 = trim(preg_replace('#[^a-z0-9]#', ' ', strtolower($this->extra)));
			$extra2 = trim(preg_replace('#[^a-z0-9]#', ' ', strtolower($parts2['extra'])));
			$extra1 = substr($extra1, 0, 32);
			$extra2 = substr($extra2, 0, 32);
			if (!strlen($extra1) || !strlen($extra2)) $score += 16;
				else $score += levenshtein($extra1, $extra2) * 3;
		}
		
		return $score;
}*/

	function getstandardname()
	{
		$parts = array();
		if ($this->surnamefirst)
		{
			$parts[] = $this->surname . ',';

			if ($this->title != '')
			{
				if ($this->givennames != '') $parts[] = $this->givennames . ',';
				$parts[] = $this->title;
			}
			else
			{
				if ($this->givennames != '') $parts[] = $this->givennames;
				$parts[] = $this->title;
			}
		}
		else
		{
			if ($this->title) $parts[] = $this->title;
			if ($this->givennames) $parts[] = $this->givennames;
			$parts[] = $this->surname;
		}
		return implode(' ', $parts);
	}

	function getmappedname($showtitle = true, $reversed = true, $showgivens = 8, $expandinitials = 0xff, $showextra = true, $surnamefirst = false)
	// $reversed will output "Smith, Tony John, Dr" instead of "Tony John Smith"
	// $surnamefirst will output "Dr Smith Tony John" instead of "Tony John Smith" (useful for cultural differences)
	// $expandinitials is a bitfield specifying which names will be expanded
	// (if we have the expanded form stored)
	// 0: "T. J. W. Smith" (no names expanded)
	// 1: "Tony J. W. Smith" (1st name expanded)
	// 2: "T. John W. Smith" (2nd name expanded)
	// 3: "Tony John W. Smith" (1st and 2nd names expanded)
	// 4: "T. J. Walter Smith" (3rd name expanded)
	// 5: "Tony J. Walter. Smith" (1st and 3rd names expanded)
	// 6: "T. John Walter. Smith" (2nd and 3rd names expanded)
	// 7: "Tony John Walter Smith" (1st, 2nd and 3nd names expanded)
	// 0xffff: "Tony John Walter Robert Malcom Jeremy Smith" (all names expanded)
	// $showgivens is the maximum number of given names to show (8 is the max)
	// $showtitle shows the title eg "Dr", "Prof"
	{
		// initialise
		$givens = $this->givennames;
		$init = $expandinitials;
		$numg = $showgivens;
		foreach ($givens as $id => $given)
		{
			if ($numg-- <= 0)
			{
				unset($givens[$id]);
				continue;
			}
			if (!($init & 1))
				$givens[$id] = $this->nametoinitial($given) . '.';
			$init >>= 1;
		}
		$textgivens = implode(' ', $givens);
		$surname = $this->surname;
		$title = $this->title;
		if ($textgivens)
		{
			$textname = $reversed ?
				"$surname, $textgivens" :
				($surnamefirst ? "$surname $textgivens" : "$textgivens $surname");
		}
		else $textname = $surname;
		if ($showtitle && $title) $textname = "$title $textname";
		$extra = $this->extra;
		if ($showextra && $extra) $textname = "$textname $extra";
		return $textname;		
	}
	
	// private
	function alpha26($name)
	{
		$name = strtr($name, "", "SOZsozYYuAAAAAAACEEEEIIIIDNOOOOOOUUUUYsaaaaaaaceeeeiiiionoooooouuuuyy");
		$name = preg_replace('#[^a-zA-Z]+#', '', $name);
		return strtoupper($name);
	}
	
	// public
	function getpid($number = 0)
	// gets a PID (unique person identifier) based on this name
	// to ensure uniqueness, call it with an increasing number (beginning 0)
	// until it returns one that is not taken
	{
		$surn = substr($this->alpha26($this->surname), 0, 4);
		$givn = substr($this->alpha26(reset($this->givennames)) ,0, 2);
		$pid = $surn . $givn;
		while (strlen($pid) < 6) $pid .= "0";
		if ($number > 0)
		{
			$numstring = "$number";
			if (strlen($numstring) > 6)
				trigger_error("Exhausted 1 million possible unique person identifiers", E_USER_ERROR);
			$pid = substr($pid, 0, 6 - strlen($numstring)) . $numstring;
		} 
		return strtoupper($pid);
	}
	
	// private
	function nametoinitial($name)
	{
		$prefixes = array('de ', 'of ', 'au ');
		foreach ($prefixes as $pref)
		{
			$patt = str_replace(' ', '\s+', $pref);
			if (preg_match("#^$patt#i", $name, $matches))
				$name = substr($name, strlen($matches[0]));
		}
		return strtoupper($name[0]);
	}
	
	// public
	function setname($surname, $givennames = null, $title = null, $extra = null)
	// name won't be translated, so it must be in normal form already, ie
	// capitalisations and periods indicating initials, etc
	{
		$this->title = $title;
		$this->surname = $surname;
		$this->givennames = preg_split('#\s+#', $givennames);		
		$this->extra = $extra;
	}
	
	// private
	function clear()
	{
		$this->surname =
			$this->givennames =
			$this->title =
			$this->extra = 
			null;
	}


	static function extract($textname)
	// returns a name object based on a name given in plain text
	{
		$nameextra = self::extractextra($textname);
		$nametitle = self::extracttitle($textname);

		// check for reversed name (surname, firstname)
		if (preg_match('#^.*[^,\s]+.*,.*[^,\s]+.*$#', $textname))
		{
			$surnamefirst = true;
			$pos = strpos($textname, ',');
			$namesurname = trim(substr($textname, 0, $pos));
			$namegivennames = trim(substr($textname, $pos+1));
			if (!$nametitle) $nametitle = self::extracttitle($namegivennames);
		}
		else
		{
			$surnamefirst = false;
			$nameparts = preg_split('#\s+#', trim($textname));
			if (empty($nameparts)) return null;
			$namesurname = array_pop($nameparts);
			while (self::isprefix(end($nameparts)))
				$namesurname = array_pop($nameparts) . " " . $namesurname;
			$namegivennames = implode(' ', $nameparts);
		}

		$namesurname = self::namemassage($namesurname);
		$namegivennames = self::namemassage($namegivennames);

		$name = new name();
		$name->surname = $namesurname;
		$name->givennames = $namegivennames;
		$name->surnamefirst = $surnamefirst;
		$name->title = $nametitle;
		$name->comment = $nameextra;

		return $name;
	}

	static function extractextra(&$textname)
	{
		// extract any extra info (eg details in brackets after the name)
		if (preg_match('#,?\s+((,?\s*(([Cc]a?\.?)?\s*\d{4}[^\[\(,]*|\[[^\]]*\]|\([^\)]*\)|i\.?\s?e\.[^\[\9,]*|[dbf]\.?\s*\d{4}\.?))+)\s*$#', 
			$textname, $matches, PREG_OFFSET_CAPTURE))
		{
			$extra = trim($matches[1][0]);
			$textname = substr($textname, 0, $matches[1][1]);
			return $extra;
		}
	}
	
	static function extracttitle(&$textname)
	{
		// extract the title (mr, mrs, ms, etc)
		$acceptedtitles = array(
			"A/Prof" => array('aprof', 'a/prof', 'apro', 'a/pro', 'associate professor', 'a/professor', 'a/p', 'a prof', 'a pro'),
			"Adj/Prof" => array('adjprof', 'adj/prof', 'adjpro', 'adj/pro', 'adjunct professor', 'adj prof', 'adj pro'),
			"Em/Prof" => array('emprof', 'em/prof', 'empro', 'em/pro', 'em prof', 'emeritus professor', 'em professor', 'professor emeritus'),
			"Prof Dr" => array('prof dr', 'professor doctor', 'professor dr', 'prof/dr'),
			"Prof" => array('prof', 'professor', 'pro'),
			"Dr" => array('dr', 'doctor', 'ph.d'),
			"Miss" => array('miss'),
			"Mr" => array('mr', 'mister'),
			"Mrs" => array('mrs', 'missus'),
			"Ms" => array('ms'),
			"Herr" => array('herr'),
			"Monsieur" => array('monsieur'),
			"Hr" => array('hr'),
			"Frau" => array('frau'),
			"Mstr" => array('mstr', 'master'),
			"Sir" => array('sir')
			);
		foreach ($acceptedtitles as $title => $permutations)
		{
			foreach ($permutations as $perm)
			{
				$patt = str_replace(' ', '\.?:?\s+', $perm);
				$patt = str_replace('.', '\.', $patt);
				if (preg_match("#^\s*$patt\.?:?\s+#i", $textname, $matches))
				{
					$textname = substr($textname, strlen($matches[0]));
					return $title;
				}
				elseif (preg_match("#,\s*$patt\.?:?\s*$#i", $textname, $matches))
				{
					$textname = substr($textname, 0, 0 - strlen($matches[0]));
					return $title;
				}
			}
		} 
		return null;
	}

	// private
	static function isprefix($name)
	{
		$prefixes = array('de', 'of', 'van', 'der', 'ten', 'von');
		if (in_array(trim(strtolower($name)), $prefixes)) return true;
		return false;
	}
	
	static function namemassage(&$textname)
	{
		$textname = trim($textname);
		$textname = preg_replace('#[()";:\[\]]++#', '', $textname);
		$textname = preg_replace('#[_\s]++#', ' ', $textname);
		return $textname;
	}

	/*
	function getalphaindex($text, $unsigned = false)
	// generates a 32 bit integer from the first few characters (max 7) of the
	// text to
	// aid sorting and fetching strings that are alphabetically nearby
	// lower numbers = earlier in the alphabet
	// if $unsigned is false, it avoids using the highest bit in case it is
	// interpreted as a signing bit (which would affect sorting)
	{
		$value = 0;
		$shift = $unsigned ? 27 : 26;
		$text = preg_replace('#[^a-z]+#', '', strtolower($text));
		$t = substr($text, 0, 7);
		for ($i = 0; $i < 7; $i++)
		{
			if ($i > strlen($t) - 1)
				$cod = 0;
			else
				$cod = ord($t{$i}) + 1 - ord('a');
			$cod = $shift >= 0 ? ($cod << $shift) : ($cod >> (0 - $shift));
			$value |= $cod;
			$shift -= 5;
		}
		return $value;
	}
	*/

}

?>
