<?php

/*
	textfuzzy - fuzzy text comparing and matching, including phoenetic matching
	Copyright (c) 2009
	
	This file is part of NicNames.
	
	NicNames is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as 
	published by the Free Software Foundation, either version 3 of
	the License, or (at your option) any later version.
	
	NicNames is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with NicNames.  If not, see <http://www.gnu.org/licenses/>.
*/

require_once(BLUESTONE_DIR . 'textnormal.inc.php');
require_once(BLUESTONE_DIR . 'diff.inc.php');

class textfuzzy
{
	function __construct($input)
	{
		$this->string = $input;
	}

	function getnormal()
	// normalised text for the purpose of natural comparison (not phonetically though)
	{
		if (isset($this->normal)) return $this->normal;

		$textnormal = new textnormal($this->string);
		return ($this->normal = $textnormal->normal(true));
	}

	function getphon($joinwords = false)
	// returns a phoenetic representation of the input text
	{
		if (isset($this->phon))
			return $joinwords ? str_replace(' ', '', $this->phon) : $this->phon;

		$text = $this->getnormal();

		static $preprocess = array(
			// common english
			'qu' => 'kw', 'kn' => 'n', 'gn' => 'n', 'lm' => 'm',
			'ld' => 'd', 'ce' => 'se', 'psy' => 'sy', 'wr' => 'r',
			'ing' => 'in', 'ich' => 'itch', 'own' => 'on', 'gua' => 'gwa',
			'igh' => 'i', 'ph' => 'f', 'cy' => 'sy', 'tion' => 'shon',
			'isch' => 'ish', 'schm' => 'shm', 'ough' => 'oe',
			'dge' => 'je', 'dgy' => 'jy', 'xx' => 'x',
			'essed' => 'est',
			// foreign words
			'aux' => 'au',
			// single letters
			'x' => 'ks',
			// digits
			'0' => 'zero', '1' => 'won', '2' => 'too', '3' => 'three', '4' => 'four', '5' => 'five',
			'6' => 'sicks', '7' => 'seven', '8' => 'eight', '9' => 'nine',
			);
		$text = strtr($text, $preprocess);

		static $regexes = array(
			'#\bone\b#' => 'won',
			'#\btwo\b#' => 'too',
			'#le\b#' => 'el',
			'#([^aeiuoy ])re\b#S' => '\\1er',
			'#([aeiouy][^aeiuoyh ]{1,3})e(?![rmnlpdstz])#S' => '\\1',
			'#(?<!^m)(?<!^ma)ci#S' => 'si',
			'#ew(?![aeiouy])#' => 'u\\1',
			//'#sh([aeiouy])#' => 'j\\1',
			'#\bmc([^aeouy ])#' => 'mac\\1',
			'#([aeiouy])w+#S' => '\\1',
			'#([aeiouy])r+([^aeiouy ])#S' => '\\1\\2',
			'#([aeiouy])r+\b#S' => '\\1',
			);
		$text = preg_replace(array_keys($regexes), array_values($regexes), $text);
		// === groups:
		// o: aeiouy
		// m: bmnp
		// k: cgkqxj
		// t: dstz
		// f: fvw
		// r: lr
		// unused: h		
		static $process = array(
			'a'=>'o','e'=>'o','i'=>'o','u'=>'o','y'=>'o',
			'b'=>'m','n'=>'m','p'=>'m',
			'c'=>'k','g'=>'k','q'=>'k','x'=>'k','j'=>'k',
			'd'=>'t','s'=>'t','z'=>'t',
			'v'=>'f','w'=>'f',
			'l'=>'r',
			'h'=>'',
			);
		$text = strtr($text, $process);
		$text = preg_replace('#([omktfr])(?=\\1)#', '', $text); // remove double letters
		$this->phon = $text;
		return $joinwords ? str_replace(' ', '', $text) : $text;
	}

	private function diffdistance($a, $b)
		// a drop-in replacement for levenshtein() which can handle long strings
		// effectively (but not recommended for short strings)
		// like levenshtein, it returns a count of bytes added, removed or changed
		// unlike levenshtein, very small matching segments may count as changed
	{
		$editlist = diff::dodiff($a, $b);
		$dist = 0;
		foreach ($editlist as $edit) $dist += max($edit[1], $edit[3]);
		return $dist;
	}

	function similarityto($other)
		// returns a score indicating the similarity or difference between this and another
		// string.
		// 1 = two strings are identical
		// 0 = there is no correlation, the two could be randomly chosen
		// 0.5 = middle ground, whether or not there is any correlation is most difficult to determine
		// if both strings are empty, it returns null.  this special case needs to be dealt with
	{
		if ($this->string == '' && $other->string == '') return null;
		if ($this->string === $other->string) return 1;
		if ($this->string == '' || $other->string == '') return 0;

		$len1 = strlen($this->string);
		$len2 = strlen($other->string);

		// for short strings, use levenshtein for comparison
		if ($len1 <= 60 && $len2 <= 60 && ($len1 <= 22 || $len2 <= 22))
			$comparefunc = 'levenshtein';
		else
			$comparefunc = array($this, 'diffdistance');	

		// first compare phoenetically
		$phon1 = $this->getphon(true);
		$phon2 = $other->getphon(true);
		$threshold = max(strlen($phon1), strlen($phon2)) * 0.6;
		$score_phon = max(0, 1 - (call_user_func($comparefunc, $phon1, $phon2) / $threshold));

		$norm1 = $this->getnormal();
		$norm2 = $other->getnormal();
		
		// compare normalised if looks near enough
		if ($score_phon > 0.5)
		{
			$threshold = max(strlen($norm1), strlen($norm2)) * 0.8;
			$score_norm = max(0, 1 - (call_user_func($comparefunc, $norm1, $norm2) / $threshold));
		}
		else $score_norm = $score_phon * 0.3;

		// take normalised into account
		// even if the normalised forms are identical, we still return less than 1
		$score = pow(($score_phon * 0.4) + ($score_norm * 0.58), 1.25);

		return $score;
	}

	function getphonkeys($reverse = false)
		// translates each word in the string into a 32-bit positive integer representing
		// it, such that similar sounding words shall have similar keys
		// returns associative array of integers, one per word in the original string
	{
		$words = explode(' ', $this->getphon());

		$results = $lookup = array();

		foreach ($words as $id => $word)
		{
			$word = substr($reverse ? strrev($word) : $word, 0, 11);
			if (isset($lookup[$word]))
			{
				$output = $lookup[$word];
			}
			else
			{
				$word = strtr($word, 'ormftk', "\x01\x02\x03\x04\x05\x06");
				$output = 0;
				for ($bit = 31, $off = 0, $strlen = strlen($word); $bit > 0 && $off < $strlen; $off++)
				{
					$val = ord($word[$off]);
					if ($val < 7)
					{	
						$output |= $bit >= 3 ? ($val << ($bit - 3)) : ($val >> (3 - $bit));	
						$bit -= 3;
					}
					else
					{
						if ($bit > 11) 
							$output |= ($val << ($bit - 11)) | (7 << ($bit - 3));
						elseif ($bit > 3)
							$output |= ($val >> (11 - $bit)) | (7 << ($bit - 3));
						else
							$output |= (7 << ($bit - 3));

						$bit -= 11;
					}
					//echo str_pad (decbin($output), 32, 0, STR_PAD_LEFT) . "\n";
				}
				$lookup[$word] = $output;
			}
			$results[] = $output;
		}
		return $results;
	}

	function getcommonwordscores()
		// returns an array of scores, one for each word in the string, indicating whether
		// the word is a common/stop word.  A very common word will get a low score like 
		// 0.2, whereas a rare or unrecognised word will 1.0.
	{
		$words = explode(' ', $this->getnormal());

		$results = array();

		foreach ($words as $id => $word)
		{
			$freq = isset(self::$commonwords[$word]) ? self::$commonwords[$word] : 0;
			$results[] = 1.0 / ($freq + 1);
		}
		return $results;
	}

	function getstemmed()
		// gets a copy of the text where all words have been stemmed.  also normalises
	{
		$text = ' ' . $this->getnormal() . ' ';

		return trim(strtr($text, array(
			// whole words
			' its ' => ' it ', ' actual ' => ' actual ', ' youre ' => ' you ', ' youll ' => ' you ',
			' arent ' => ' are ', ' dont ' => ' do ', ' ring ' => ' ring ', ' king ' => ' king ',
			' ding ' => ' ding ', ' native ' => ' native ', ' ties ' => ' tie ', ' untie ' => ' untie ',
			' unties ' => ' untie ', ' ping ' => ' ping ', ' these ' => ' this ', ' theyre ' => ' they ',
			' under ' => ' under ', ' motion ' => ' motion ', ' says ' => ' say ',
			// 5m
			'ntation ' => 'nt ', 'llation ' => 'll ', 'nstrably ' => 'nstrate ', 'ddition ' => 'dd ',
			'dditional ' => 'dd ', 'isation ' => 'ise ', 'ization ' => 'ise ',
			'nning ' => 'n ', 'nities ' => 'ne ', 'mming ' => 'm ', 'netic ' => 'neme ', 'nding ' => 'nd ',
			'erent ' => 'er ', 'ctions ' => 'ct ', 'going ' => 'go ', 'dding ' => 'dd ', 'ptions ' => 'pt ',
			'tation ' => 'te ', 'lication ' => 'ly ', 'tigation ' => 't ', 'plied ' => 'ply ',
			'rcial ' => 'rce ', 'ional ' => 'ion ', 'rtant ' => 'rt ', 'ntific ' => 'nce ', 'liance ' => 'ly ',
			'pment ' => 'p ', 'cement ' => 'ce ', 'wledge ' => 'w ', 'riety ' => 'ry ', 'emmer ' => 'em ',
			'osing ' => 'ose ', 'utions ' => 'ute ', 'uding ' => 'ude ', 'ogist ' => 'ogy ', 'aking ' => 'ake ',
			'ouses ' => 'ouse ', 'ative ' => 'ate ', 'ogical ' => 'ogy ', 'being ' => 'be ',
			'pings ' => 'p ', 'pirical ' => 'pire ', 'vable ' => 've ', 'omena ' => 'omenon ', 'idity ' => 'id ',
			'chers ' => 'ch ', 'linary ' => 'line ', 'alized ' => 'alise ', 'alised ' => 'alise ',
			'ogies ' => 'ogy ', 'cracy ' => 'crat ', 'cratic ' => 'crat ', 'alist ' => 'al ', 'alists ' => 'al ',
			'vement ' => 've ', 'vements ' => 've ', 'merly ' => 'mer ', 'aimed ' => 'aim ', 'adian ' => 'ada ',
			'alian ' => 'alia ', 'nder ' => 'nd ', 'estern ' => 'est ', 'ident ' => 'idence ',
			'alise ' => 'al ', 'rtise ' => 'rt ', 'alized ' => 'al ', 'alised ' => 'al ', 'alize ' => 'al ',
			'rrences ' => 'r ', 'lves ' => 'lf ', 'motion ' => 'mote ', 'ntial ' => 'nce ',
			'ntage ' => 'nt ', 'loped ' => 'lop ', 'assed ' => 'ass ', 'apita ' => 'apital ', 'isses ' => 'iss ',
			'ating ' => 'ate ', 'riable ' => 'ry ', 'riables ' => 'ry ', 'stence ' => 'st ', 'ithin ' => 'ith ',
			'mbling ' => 'mble ', 'onely ' => 'one ', 'ptions ' => 'ption ', 'inition ' => 'ine ',
			'oader '=> 'oad ',
			// 4
			'nity ' => 'ne ', 'king ' => 'k ', 'sily ' => 'se ', 'nted ' => 'nt ', 'ries ' => 'ry ',
			'nded ' => 'nd ', 'fied ' => 'fy ', 'ines ' => 'ine ', 'ered ' => 'er ', 'tric ' => 'try ',
			'ures ' => 'ure ', 'aged ' => 'age ', 'thms ' => 'thm ', 'vity ' => 've ', 'tual ' => 't ',
			'mary ' => 'me ', 'ring ' => 'r ', 'ting ' => 't ', 'uses ' => 'us ', 'sues ' => 'sue ',
			'mmed ' => 'm ', 'ding ' => 'd ', 'ches ' => 'ch ', 'ores ' => 'ore ', 'nces ' => 'nce ',
			'ties ' => 'ty ', 'cher ' => 'ch ', 'ated ' => 'ate ', 'uped ' => 'up ', 'ably ' => 'able ',
			'ives ' => 'ive ', 'ical ' => 'ic ', 'ions ' => 'ion ', 'eces ' => 'ece ', 'mies ' => 'my ',
			'cies ' => 'cy ', 'ides ' => 'ide ', 'ails ' => 'ail ', 'rred ' => 'r ', 'dler ' => 'dle ',
			'fies ' => 'fy ', 'sted ' => 'st ', 'aces ' => 'ace ', 'kies ' => 'kie ', 'phic ' => 'ph ',
			'nues ' => 'nue ', 'ator ' => 'ate ', 'omes ' => 'ome ', 'unny ' => 'un ', 'ites ' => 'ite ',
			'nths ' => 'nth ',
			// 3
			'ses ' => 'se ', 'lly ' => 'l ', 'les ' => 'le ', 'sed ' => 'se ',
			'nal ' => 'n ', 'ked ' => 'k ', 'ics ' => 'ic ', 'ned ' => 'ne ',
			'ems ' => 'em ', 'rns ' => 'rn ', 'ubs ' => 'ub ', 'ans ' => 'an ',
			'nly ' => 'n ', 'rms ' => 'rm ', 'ews ' => 'ew ', 'als ' => 'al ',
			'ows ' => 'ow ', 'xes ' => 'x ', 'ops ' => 'op ', 'nes ' => 'ne ',
			// 2
			'ts ' => 't ', 'rs ' => 'r ', 'ds ' => 'd ', 'ns ' => 'n ',
			// 1
			// US-british
			'iz' => 'is', 'color' => 'colour', 'eled ' => 'elled ',
			)));

		/*
		// todo the following are based on porter's algorithm - may still contain some ideas
		static $stemrules = array(
			// step 1
			'/sses\b/' => 'ss',
			'/ies\b/' => 'i',
			'/(?<!s)s\b/S' => '',
			'/eed\b/' => 'ee',
			'/at(?:ing|ed)\b/' => 'ate',
			'/bl(?:ing|ed)\b/' => 'ble',
			'/iz(?:ing|ed)\b/' => 'ize',
			'/([^lsz ])\\1(?:ing|ed)\b/' => '\\1',
			'/([^aeiouy ][aeiouy][^aeiouywx ])(?:ing|ed)\b/' => '\\1e',
			'/[aeiouy][^ ]+y\b/' => 'i',
			// step 2
			'/ational\b/' => 'ate',
			'/tional\b/' => 'tion',
			'/([ae])nci\b/' => '\\1nce',
			'/i[sz]er\b/' => 'ize',
			'/logy\b/' => 'log',
			'/entli\b/' => 'ent',
			'/ousli\b/' => 'ous',
			'/alli\b/' => 'al',
			'/bli\b/' => 'ble',
			'/eli\b/' => 'e',
			'/i[sz]ation\b/' => 'ize',
			'/at(?:ion|or)\b/' => 'ate',
			'/iveness\b/' => 'ive',
			'/fulness\b/' => 'ful',
			'/ousness\b/' => 'ous',
			'/alism\b/' => 'al',
			'/biliti\b/' => 'ble',
			'/aliti\b/' => 'al',
			'/iviti\b/' => 'ive',
			// step 3
			'/ic(?:al|ate|iti)\b/' => 'ic',
			'/ness\b|ful\b|ative\b/' => '',
			'/ali[sz]e\b/' => 'al',
			// step 4
			'/al\b|ance\b|ence\b|er\b|ic\b|able\b|ible\b|ant\b|ement\b|ment\b|ent\b|ism\b|ate\b|iti\b|ous\b|ive\b|ize\b|ou\b/' => '',
			'/([st])ion\b/' => '\\1',
			// step 5
			'/([^aeiouy ]++[aeiouy]++[^aeiouy]++)e\b/' => '\\1',
			'/([^aeiouy ]++[aeiouy]++[^aeiouy]+?([^aeiouy])\\2)l\b/' => '\\1',
			);
		return preg_replace(array_keys($stemrules), array_values($stemrules), $text);
		 */
	}

	static $commonwords = array(
		// english
		'the'=>5,'of'=>5,'to'=>5,'and'=>5,'a'=>5,'in'=>5,'is'=>5,'it'=>5,'you'=>5,
		'that'=>5,'he'=>5,'was'=>5,'for'=>5,'on'=>5, 'are'=>5,'with'=>5,'as'=>5,
		'i'=>4,'his'=>4,'they'=>4,'be'=>4,'at'=>4,'one'=>4,'have'=>4,'this'=>4,
		'from'=>4, 'or'=>4,'had'=>4,'by'=>4,'hot'=>4,'word'=>4,'but'=>4,'what'=>4,
		'some'=>4,'we'=>4,'can'=>4,'out'=>4,'other'=>4, 'were'=>4,'all'=>4,
		'there'=>3,'when'=>3,'up'=>3,'use'=>3,'your'=>3,'how'=>3,'said'=>3,
		'an'=>3,'each'=>3, 'she'=>3,'which'=>3,'do'=>3,'their'=>3,'time'=>3,
		'if'=>3,'will'=>3,'way'=>3,'about'=>3,'many'=>3,'then'=>3, 'them'=>3,
		'write'=>2,'would'=>2,'like'=>2,'so'=>2,'these'=>2,'her'=>2,'long'=>2,
		'make'=>2,'thing'=>2,'see'=>2, 'him'=>2,'two'=>2,'has'=>2,'look'=>2,
		'more'=>2,'day'=>2,'could'=>2,'go'=>2,'come'=>2,'did'=>2,'number'=>2,
		'sound'=>2,'no'=>2,'most'=>2,'people'=>2,'my'=>2,'over'=>2,'know'=>2,
		'water'=>1,'than'=>1,'call'=>1, 'first'=>1,'who'=>1,'may'=>1,'down'=>1,
		'side'=>1,'been'=>1,'now'=>1,'find'=>1,'any'=>1,'new'=>1,'work'=>1,
		'part'=>1,'take'=>1,'get'=>1,'place'=>1,'made'=>1,'live'=>1,'where'=>1,
		'after'=>1,'back'=>1,'little'=>1, 'only'=>1,'round'=>1,'man'=>1,'year'=>1,
		'came'=>1,'show'=>1,'every'=>1,'good'=>1,'me'=>1,'give'=>1,'our'=>1,
		// german
		'der'=>2,'die'=>2,'und'=>2,'in'=>2,'den'=>2,'von'=>1,'zu'=>1,'das'=>1,
		'mit'=>1,'sich'=>1,'des'=>1,'auf'=>1,
		// french
		'le'=>2,'de'=>2,'un'=>2,'etre'=>2,'et'=>2,'il'=>1,'avoir'=>1,
		'ne'=>1,'je'=>1,'son'=>1,'que'=>1,'se'=>1,
		// spanish
		'de'=>2,'el'=>2,'que'=>2,'la'=>2,'en'=>2,'y'=>1,'los'=>1,'se'=>1,
		'ser'=>1,'un'=>1,'por'=>1,'con'=>1,
		);



/*
	// private
	function alpha26($name)
	{
		trigger_error('need to do textnormal to get only 26 alphas');
	}
	
	// private
	function nametoinitial($name)
	{
		$prefixes = array('de ', 'of ', 'au ');
		foreach ($prefixes as $pref)
		{
			$patt = str_replace(' ', '\s+', $pref);
			if (preg_match("#^$patt#i", $name, $matches))
				$name = substr($name, strlen($matches[0]));
		}
		return strtoupper($name[0]);
	}
	
	// public
	function setname($surname, $givennames = null, $title = null, $extra = null)
	// name won't be translated, so it must be in normal form already, ie
	// capitalisations and periods indicating initials, etc
	{
		$this->title = $title;
		$this->surname = $surname;
		$this->givennames = preg_split('#\s+#', $givennames);		
		$this->extra = $extra;
	}
	
	// private
	function clear()
	{
		$this->surname =
			$this->givennames =
			$this->title =
			$this->extra = 
			null;
	}
	
	// public
	function extract($textname)
	// extracts a name from a name given in plain text
	{
		$this->clear();
		
		$textname = $this->extractextra($textname);
		$textname = $this->extracttitle($textname);
				
		// do reversal
		$reversalparts = preg_split('#\s*,\s*#', $textname);
		foreach ($reversalparts as $id => $val) $reversalparts[$id] = trim($val);
		$textname = implode(" ", array_reverse($reversalparts));
		// name is now in forward order
		if (!$this->title) $textname = $this->extracttitle($textname);
		$nameparts = preg_split('#\s+#', trim($textname));
		if (!count($nameparts)) return;
		$surname = array_pop($nameparts);
		while ($this->isprefix(end($nameparts)))
			$surname = array_pop($nameparts) . " " . $surname;
		$this->namemassage($surname);
		$this->surname = $surname;
		array_walk($nameparts, array($this, 'namemassage'));
		$this->givennames = $nameparts; 		
	}
	
	// private
	function isprefix($name)
	{
		$prefixes = array('de', 'of', 'van', 'der', 'ten');
		if (in_array(trim(strtolower($name)), $prefixes)) return true;
		return false;
	}
	
	// private
	function namemassage(&$textname)
	{
		$textname = trim($textname);
		$textname = preg_replace('#[()";:\[\]]#', '', $textname);
		$textname = preg_replace('#[_\s]+#', ' ', $textname);
		if (preg_match('#^(\w\.|[b-hj-tv-xz])$#', $textname))
		{
			// initialisation
			$textname = strtoupper($textname[0] . '.');
			return;
		}
		$textname = str_replace('.', '', $textname);
		$textname = strtolower($textname);
		if (preg_match_all('#(?<=^|\b|^mc|^mac)\w#i', $textname, $matches, PREG_OFFSET_CAPTURE))
		{
			foreach ($matches[0] as $val)
			{
				$offset = $val[1];
				list($word) = explode(' ', substr($textname, $offset));
				if (!$this->isprefix($word))
					$textname = substr($textname, 0, $offset) . strtoupper($textname[$offset]) . substr($textname, $offset+1);
			}
		}
		return;
	}
	
	// private
	function extractextra($textname)
	// gets extra information from the end of a name such as date of birth/death,
	// information in brackets etc
	// returns the $textname with the extra information removed, if any was
	// found
	// \s*-?\s*(\d{2,4}|or\s+\d{2,4})?\.?
	{
		if (preg_match('#,?\s+((,?\s*(([Cc]a?\.?)?\s*\d{4}[^\[\(,]*|\[[^\]]*\]|\([^\)]*\)|i\.?\s?e\.[^\[\9,]*|[dbf]\.?\s*\d{4}\.?))+)\s*$#', $textname, $matches, PREG_OFFSET_CAPTURE))
		{
			$this->extra = trim($matches[1][0]);
			$textname = substr($textname, 0, $matches[1][1]);
		}
		
		return $textname;
	}
	
	
	function getindexes()
	// gets phonetic indexes in associative array
	{
		$sur_phon = $this->getphon($this->surname);
		$indexes = array();
		$indexes['i_surn'] = $this->getalphaindex($this->surname);
		$indexes['i_given'] = $this->getalphaindex(implode(' ', $this->givennames));
		$indexes['i_phon'] = $this->getalphaindex($sur_phon);
		$indexes['i_phon_r'] = $this->getalphaindex(strrev($sur_phon));
		return $indexes;		
	}
	
		
	function compareto($other)
	// calculates score indicating how different this name is to the other
	// takes similar spelling and pronunciation into account
	// the closer to zero, the surer this algorithm is that the names are the same
	// this algorithm is fairly processor intensive
	// 0-22: most likely the same name eg same first name and surname
	// 23-32: looks like same name eg same surname and first initial
	// 33-50: could be the same except for spelling or small variations 
	// 51-100: doesn't look like a match, could be relevant though
	// 101-: doesn't look relevant 
	{
		
		// surname
		$dif = $this->comparestrings($this->surname, $parts2['surname']);
		$bon = in_array($parts2['surname'], $this->givennames)
			||   in_array($this->surname, $parts2['givennames']);
		$score = $bon ? min(40, $dif) : $dif;		
		
		// givennames
		$given1 = $this->givennames;
		$given2 = $parts2['givennames'];
		$scorevals = array(50, 30, 10, 8, 6, 4, 2, 1);
		while (!empty($scorevals))
		{
			$name1 = array_shift($given1);
			$name2 = array_shift($given2);
			$scoreval = array_shift($scorevals);
			$nscore = 0;
			if (!strlen($name1) || !strlen($name2)) $nscore = 35;
			else
			{
				$initials = (preg_match('#\.$#', $name1) || preg_match('#\.$#', $name2));				if (!$initials)
					$nscore = $this->comparestrings($name1, $name2);
				else
					$nscore = 20;	
				if ($name1{0} != $name2{0}) $nscore += 50;
			}
			$score += round(($nscore * $scoreval) / 100);
		}
		
		// title
		if ($this->title != $parts2['title'])
		{
			if (!strlen($this->title) || !strlen($parts2['title'])) $score += 2;
				else $score += 6;
		}
		
		// extra
		if ($this->extra != $parts2['extra'])
		{
			$extra1 = trim(preg_replace('#[^a-z0-9]#', ' ', strtolower($this->extra)));
			$extra2 = trim(preg_replace('#[^a-z0-9]#', ' ', strtolower($parts2['extra'])));
			$extra1 = substr($extra1, 0, 32);
			$extra2 = substr($extra2, 0, 32);
			if (!strlen($extra1) || !strlen($extra2)) $score += 16;
				else $score += levenshtein($extra1, $extra2) * 3;
		}
		
		return $score;

	}
	
	function getalphaindex($text, $unsigned = false)
	// generates a 32 bit integer from the first few characters (max 7) of the
	// text to
	// aid sorting and fetching strings that are alphabetically nearby
	// lower numbers = earlier in the alphabet
	// if $unsigned is false, it avoids using the highest bit in case it is
	// interpreted as a signing bit (which would affect sorting)
	{
		$value = 0;
		$shift = $unsigned ? 27 : 26;
		$text = preg_replace('#[^a-z]+#', '', strtolower($text));
		$t = substr($text, 0, 7);
		for ($i = 0; $i < 7; $i++)
		{
			if ($i > strlen($t) - 1)
				$cod = 0;
			else
				$cod = ord($t{$i}) + 1 - ord('a');
			$cod = $shift >= 0 ? ($cod << $shift) : ($cod >> (0 - $shift));
			$value |= $cod;
			$shift -= 5;
		}
		return $value;
	}

	private function extracttitle($textname)
	// gets the "title" from the start of the text based name ie Ms, Dr, etc
	// returns the name - with the title removed if one was found
	{
		$acceptedtitles = array(
			"A/Prof" => array('aprof', 'a/prof', 'apro', 'a/pro', 'associate professor', 'a/professor', 'a/p', 'a prof', 'a pro'),
			"Adj/Prof" => array('adjprof', 'adj/prof', 'adjpro', 'adj/pro', 'adjunct professor', 'adj prof', 'adj pro'),
			"Em/Prof" => array('emprof', 'em/prof', 'empro', 'em/pro', 'em prof', 'emeritus professor', 'em professor', 'professor emeritus'),
			"Prof Dr" => array('prof dr', 'professor doctor', 'professor dr', 'prof/dr'),
			"Prof" => array('prof', 'professor', 'pro'),
			"Dr" => array('dr', 'doctor', 'ph.d'),
			"Miss" => array('miss'),
			"Mr" => array('mr', 'mister'),
			"Mrs" => array('mrs', 'missus'),
			"Ms" => array('ms'),
			"Herr" => array('herr'),
			"Monsieur" => array('monsieur'),
			"Hr" => array('hr'),
			"Frau" => array('frau'),
			"Mstr" => array('mstr', 'master'),
			"Sir" => array('sir')
			);
		foreach ($acceptedtitles as $title => $permutations)
		{
			foreach ($permutations as $perm)
			{
				$patt = str_replace(' ', '\.?:?\s+', $perm);
				$patt = str_replace('.', '\.', $patt);
				if (preg_match("#^\s*$patt\.?:?\s+#i", $textname, $matches))
				{
					$textname = substr($textname, strlen($matches[0]));
					$this->title = $title;
					return $textname;
				}
			}
		} 
		return $textname;
	}	
	 */

}

?>
